/* /////////////////////////////////////////////////////////////////////////////////////////////////
ConnectAudio v0.3 by MacGruber.
Control E-Stim devices via external application.

Version 0.3 - 2021-02-27
	Fixed issues caused by VaM's new security measures, works with VaM 1.20.77.9 now.

Version 0.2 - 2020-10-11
	(no code changes)

Version 0.1 - 2020-09-26
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace MacGruber.Connect
{
	public class Audio : MVRScript
	{	
		private Socket myClient;	
		private IAsyncResult myResult;		
		private IPAddress myIpAddress;
		private DateTime myLastSendTimestamp = DateTime.MinValue;
		private bool myConnected = false;
		
		
		private JSONStorableString myCommand;
		
		private const bool USE_IPv6 = false;
		private readonly string HOSTNAME = "localhost";
		private const int PORT = 39340;
		private readonly TimeSpan ALIVE_CHECK_TIME = new TimeSpan(0, 0, 1);
		
		private readonly byte[] myBuffer = new byte[256];
				
		public override void Init()
		{	
			myCommand = new JSONStorableString("Send", string.Empty, Send);
			RegisterString(myCommand);
		}
		
		private void Update()
		{
			try
			{
				if (myClient == null)
				{
					myConnected = false;
					AddressFamily mode = USE_IPv6 ? AddressFamily.InterNetworkV6 : AddressFamily.InterNetwork;	
					IPHostEntry hostEntry = Dns.GetHostEntry(HOSTNAME);
					myIpAddress = Array.Find(hostEntry.AddressList, ip => ip.AddressFamily == mode);
					LogMessage("Trying to connect to {0}:{1}...", myIpAddress, PORT);
					myClient = new Socket(myIpAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
					myClient.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.KeepAlive, true);
					myResult = myClient.BeginConnect(new IPEndPoint(myIpAddress, PORT), null, null);
				}
				else if (myConnected)
				{
					if (!myClient.Connected)
					{
						myConnected = false;
						myClient.Close();
						myClient = null;
						LogMessage("Disconnected.");
					}
					else 
					{
						if (myLastSendTimestamp + ALIVE_CHECK_TIME < DateTime.Now)
						{
							// send message of length 0 to check if the connection is still there
							myBuffer[0] = 0;
							myClient.Send(myBuffer, 1, SocketFlags.None);
							myLastSendTimestamp = DateTime.Now;
						}
					}
				}
				else if (myResult == null)
				{
					myResult = myClient.BeginConnect(new IPEndPoint(myIpAddress, PORT), null, null);
				}			
				else if (myResult.AsyncWaitHandle.WaitOne(0, true))
				{	
					if (myClient.Connected)	
					{
						LogMessage("Connected!");
						myClient.EndConnect(myResult);
						myLastSendTimestamp = DateTime.Now;
						myConnected = true;
					}
					myResult = null;
				}
			}
			catch (SocketException e)
			{
				string msg = e.Message.TrimEnd('\n');
				LogMessage("{0}: {1}", e.SocketErrorCode, msg);
			}
			catch (Exception e)
			{
				LogError("Error has occured: " + e.ToString());
			}
		}
					
		private void OnDisable()
		{
			try
			{
				if (myClient != null)
				{
					if (myClient.Connected)
						myClient.Shutdown(SocketShutdown.Both);  
					myClient.Close();
					myClient = null;
					LogMessage("Disconnected.");
				}
			}
			catch (SocketException e)
			{
				string msg = e.Message.TrimEnd('\n');
				LogMessage("{0}: {1}", e.SocketErrorCode, msg);
			}
			catch (Exception e)
			{
				LogError("Error has occured: " + e.ToString());
			}
		}
			
		private void Send(string message)
		{		
			try
			{			
				myCommand.valNoCallback = string.Empty;		
				if (myClient == null || !myConnected)
					return;			
				if (message.Length > 255)
				{
					LogMessage("Message too long! (max 255 characters)");
					return;
				}	
				int written = Encoding.ASCII.GetBytes(message, 0, message.Length, myBuffer, 1);
				// first byte indicates length of the message
				myBuffer[0] = (byte)written;
				myClient.Send(myBuffer, 1+written, SocketFlags.None);
				myLastSendTimestamp = DateTime.Now;
			}
			catch (SocketException e)
			{
				string msg = e.Message.TrimEnd('\n');
				LogMessage("{0}: {1}", e.SocketErrorCode, msg);
			}
			catch (Exception e)
			{
				LogMessage("Error has occured: " + e.ToString());
			}
		}
		
		private static void LogError(string msg)
		{
			SuperController.LogError("Connect.Audio: " + msg);
		}				
	
		private static void LogMessage(string msg)
		{
			SuperController.LogMessage("Connect.Audio: " + msg);
		}
		
		private static void LogMessage(string msg, params object[] args)
		{
			SuperController.LogMessage("Connect.Audio: " + string.Format(msg, args));
		}
	}
}