﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Collections;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using SimpleJSON;

namespace MacGruber
{
	// Collection of various utility functions
    public static class Utils
    {
		// VaM Plugins can contain multiple Scripts, if you load them via a *.cslist file. This function allows you to get
		// an instance of another script within the same plugin, allowing you directly interact with it by reading/writing
		// data, calling functions, etc.
		public static T FindWithinSamePlugin<T>(MVRScript self) where T : MVRScript
		{
			int i = self.name.IndexOf('_');
			if (i < 0)
				return null;
			string prefix = self.name.Substring(0, i+1);
			string scriptName = prefix + typeof(T).FullName;	
			return self.containingAtom.GetStorableByID(scriptName) as T;
		}
		
		// Get spawned prefab from CustomUnityAsset atom. Note that these are loaded asynchronously,
		// this function returns null while the prefab is not yet there.
		public static GameObject GetCustomUnityAsset(Atom atom, string prefabName)
		{
			Transform t = atom.transform.Find("reParentObject/object/rescaleObject/"+prefabName+"(Clone)");
			if (t == null)
				return null;
			else
				return t.gameObject;
		}		
		
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
		
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
		
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		// ===========================================================================================
		
		// Create VaM-UI Toggle button
		public static JSONStorableBool SetupToggle(MVRScript script, string label, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(label, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateToggle(storable, rightSide);
			script.RegisterBool(storable);
			return storable;
		}
		
		// Create VaM-UI Float slider
		public static JSONStorableFloat SetupSliderFloat(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateSlider(storable, rightSide);
			script.RegisterFloat(storable);
			return storable;
		}
		
		// Create VaM-UI ColorPicker
		public static JSONStorableColor SetupColor(MVRScript script, string label, Color color, bool rightSide)
		{
			HSVColor hsvColor = HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
			JSONStorableColor storable = new JSONStorableColor(label, hsvColor);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateColorPicker(storable, rightSide);
			script.RegisterColor(storable);
			return storable;
		}
		
		// Create VaM-UI StringChooser
		public static JSONStorableStringChooser SetupStringChooser(MVRScript self, string label, List<string> entries, bool rightSide)
		{
			string defaultEntry = entries.Count > 0 ? entries[0] : "";
			JSONStorableStringChooser storable = new JSONStorableStringChooser(label, entries, defaultEntry, label);
			self.CreateScrollablePopup(storable, rightSide);
			self.RegisterStringChooser(storable);
			return storable;
		}
		
		// Create VaM-UI StringChooser
		public static JSONStorableStringChooser SetupStringChooser(MVRScript self, string label, List<string> entries, int defaultIndex, bool rightSide)
		{
			string defaultEntry = (defaultIndex >= 0 && defaultIndex < entries.Count) ? entries[defaultIndex] : "";
			JSONStorableStringChooser storable = new JSONStorableStringChooser(label, entries, defaultEntry, label);
			self.CreateScrollablePopup(storable, rightSide);
			self.RegisterStringChooser(storable);
			return storable;
		}
		
		// Create VaM-UI StringChooser for Enum
		public static JSONStorableStringChooser SetupEnumChooser<TEnum>(MVRScript self, string label, TEnum defaultValue, bool rightSide, EnumSetCallback<TEnum> callback)
			where TEnum : struct, IComparable, IConvertible, IFormattable
		{
			List<string> names = Enum.GetNames(typeof(TEnum)).ToList();			
			JSONStorableStringChooser storable = new JSONStorableStringChooser(label, names, defaultValue.ToString(), label);
			storable.setCallbackFunction += (string name) => {
				TEnum v = (TEnum)Enum.Parse(typeof(TEnum), name);
				callback(v);
			};
			self.CreateScrollablePopup(storable, rightSide);
			self.RegisterStringChooser(storable);
			return storable;
		}
		
		// Create VaM-UI TextureChooser. Note that you are responsible for destroying the texture when you don't need it anymore.
		public static JSONStorableUrl SetupTexture2DChooser(MVRScript self, string label, string defaultValue, bool rightSide, TextureSettings settings, TextureSetCallback callback)
		{
			JSONStorableUrl storable = new JSONStorableUrl(label, string.Empty, (string url) => { QueueLoadTexture(url, settings, callback); }, "jpg|png|tif|tiff");
			self.RegisterUrl(storable);
			UIDynamicButton button = self.CreateButton("Browse " + label, false);			
			UIDynamicTextField textfield = self.CreateTextField(storable, false);
			textfield.UItext.alignment = TextAnchor.MiddleRight;
			textfield.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
			textfield.UItext.verticalOverflow = VerticalWrapMode.Truncate;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
			if (!string.IsNullOrEmpty(defaultValue))
				storable.SetFilePath(defaultValue);
			storable.RegisterFileBrowseButton(button.button);
			return storable;
		}
		
		// Create VaM-UI AssetBundleChooser.
		public static JSONStorableUrl SetupAssetBundleChooser(MVRScript self, string label, string defaultValue, bool rightSide, string fileExtensions)
		{
			JSONStorableUrl storable = new JSONStorableUrl(label, defaultValue, fileExtensions);
			self.RegisterUrl(storable);
			UIDynamicButton button = self.CreateButton("Select " + label, false);
			UIDynamicTextField textfield = self.CreateTextField(storable, false);
			textfield.UItext.alignment = TextAnchor.MiddleRight;
			textfield.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
			textfield.UItext.verticalOverflow = VerticalWrapMode.Truncate;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
			if (!string.IsNullOrEmpty(defaultValue))
				storable.SetFilePath(defaultValue);
			storable.RegisterFileBrowseButton(button.button);
			return storable;
		}	
			
		// Create VaM-UI InfoText field
		public static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
		
		// Create VaM-UI InfoText field
		public static JSONStorableString SetupInfoOneLine(MVRScript script, string text, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = script.CreateTextField(storable, rightSide);
			textfield.UItext.alignment = TextAnchor.LowerLeft;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
			return storable;
		}
		
		// Create VaM-UI button
		public static void SetupButton(MVRScript script, string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = script.CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
		}
				
		// Create input action trigger
		public static JSONStorableAction SetupAction(MVRScript script, string name, JSONStorableAction.ActionCallback callback)
		{
			JSONStorableAction action = new JSONStorableAction(name, callback);
			script.RegisterAction(action);
			return action;
		}

		// ===========================================================================================

		// Helper to add a component if missing.
		public static T GetOrAddComponent<T>(Component c) where T : Component
		{
			T t = c.GetComponent<T>();
			if (t == null)
				t = c.gameObject.AddComponent<T>();
			return t;
		}
			
		// ===========================================================================================
				
		private static void QueueLoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
		{ 
			if (ImageLoaderThreaded.singleton == null)
				return;
			if (string.IsNullOrEmpty(url))
				return;
			
			ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
			queuedImage.imgPath = url;
			queuedImage.forceReload = true;
			queuedImage.skipCache = true;
			queuedImage.compress = settings.compress;
			queuedImage.createMipMaps = settings.createMipMaps;
			queuedImage.isNormalMap = settings.isNormalMap;			
			queuedImage.linear = settings.linearColor;
			queuedImage.createAlphaFromGrayscale = settings.createAlphaFromGrayscale;
			queuedImage.createNormalFromBump = settings.createNormalFromBump;
			queuedImage.bumpStrength = settings.bumpStrength;
			queuedImage.isThumbnail = false;
			queuedImage.fillBackground = false;
			queuedImage.invert = false;
			queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
			{
				Texture2D tex = qi.tex;
				if (tex != null)
				{
					tex.wrapMode = settings.wrapMode;
					tex.filterMode = settings.filterMode;
					tex.anisoLevel = settings.anisoLevel;
				}				
				callback(tex);
			};
			ImageLoaderThreaded.singleton.QueueImage(queuedImage);			
		}
	}
	
	// ===========================================================================================
	
	public delegate void EnumSetCallback<TEnum>(TEnum v);
	public delegate void TextureSetCallback(Texture2D tex);
	
	public class TextureSettings
	{
		public bool compress = false;
		public bool createMipMaps = true;
		public bool isNormalMap = false;
		public bool linearColor = true; // Using linear or sRGB color space.
		public bool createAlphaFromGrayscale = false;
		public bool createNormalFromBump = false;
		public float bumpStrength = 1.0f;
		public TextureWrapMode wrapMode = TextureWrapMode.Repeat;
		public FilterMode filterMode = FilterMode.Trilinear;
		public int anisoLevel = 5; // 0: Forced off, 1: Off, quality setting can override, 2-9: Anisotropic filtering levels.
	}
	
	// ===========================================================================================
	
	public class AssetBundleAudioClip : NamedAudioClip
	{
		public AssetBundleAudioClip(Request aRequest, string aPath, string aName)
		{
			manager = null;
			sourceClip = aRequest.assetBundle?.LoadAsset<AudioClip>(aPath + aName);
			uid = aName;
			displayName = aName;
			category = "AssetBundle";
			destroyed = false;
		}
	}
	
	// ===========================================================================================
	
	// TriggerHandler implementation for easier handling of custom triggers.
	// Essentially call this in your plugin init code:
	//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
	//
	// Credit to AcidBubbles for figuring out how to do custom triggers.
	public class SimpleTriggerHandler : TriggerHandler
	{
		public static bool Loaded { get; private set; }
		
		private static SimpleTriggerHandler myInstance;	
		
		private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;
		
		public static SimpleTriggerHandler Instance {
			get {
				if (myInstance == null)
					myInstance = new SimpleTriggerHandler();
				return myInstance;
			}
		}
		
		public static IEnumerator LoadAssets()
		{
			yield return Instance.LoadAssetsInternal();
		}
		
        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
				yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
				yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
				yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
				yield return x;
			
			Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
			assign(prefab);
        }
		
		
		void TriggerHandler.RemoveTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {	
			return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
	}
	
	// Wrapper for easier handling of custom triggers.
	public class EventTrigger : Trigger
	{
		public string Name { get; private set; }
		public MVRScript Owner  { get; private set; }
		private bool myNeedInit = true;
		
		public EventTrigger(MVRScript owner, string name)
		{
			Name = name;
			Owner = owner;
			handler = SimpleTriggerHandler.Instance;
		}
		
		public void OpenPanel()
		{
			if (!SimpleTriggerHandler.Loaded)
				return;
			
			triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
			if (myNeedInit)
			{
				triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
				triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
				triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
				triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
				myNeedInit = false;
			}
		}
		
		public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
		{
			if (jc.HasKey(Name))
			{
				JSONClass tc = jc[Name].AsObject;
				if (tc != null)
					base.RestoreFromJSON(tc, subScenePrefix, isMerge);
			}
			else if (setMissingToDefault)
			{
				base.RestoreFromJSON(new JSONClass());
			}	
		}
		
		public void Trigger()
		{
			active = true;
			active = false;
		}
	}
}
